# Copyright (C) 2025 Embedl AB

"""Save a model after tuning."""

from pathlib import Path
from tempfile import TemporaryDirectory
from typing import List

import torch
from torch import nn

from embedl_hub.core.hub_logging import console
from embedl_hub.core.tuning.load_model import clean_model_id


def _save_jit_to_path(
    model: nn.Module,
    model_id: str,
    image_size: List[int],
    save_path: Path,
) -> Path:
    example_data = torch.randn(1, 3, *image_size)
    clean_model_name = clean_model_id(model_id)

    saved_model_path = save_path / f"{clean_model_name}_tuned.pt"

    torch.jit.save(
        torch.jit.script(model, example_inputs=[(example_data,)]),
        saved_model_path,
    )
    return saved_model_path


def save_jit_model(
    model: nn.Module, model_id: str, image_size: List[int]
) -> None:
    """Save a model in current working directory using TorchScript."""
    saved_model_path = _save_jit_to_path(
        model,
        model_id,
        image_size,
        Path.cwd(),
    )
    console.print(
        f"[green]✓ Model {model_id} saved as TorchScript model at {saved_model_path}[/]"
    )


def assert_can_save_jit_model(
    model: nn.Module, model_id: str, image_size: List[int]
) -> None:
    """Assert that the model can be saved as a TorchScript model."""
    with TemporaryDirectory() as temp_dir:
        try:
            _save_jit_to_path(model, model_id, image_size, Path(temp_dir))
        except Exception as e:
            raise RuntimeError(
                f"Model {model_id} cannot be saved as a TorchScript model. "
                f"Error: {e}"
            ) from e



# Copyright (C) 2025 Embedl AB

"""
Public Embedl Hub library API.
```pycon
>>> import embedl_hub
>>> embedl_hub.__version__
'2025.9.0'
```
"""

import importlib.metadata

from embedl_hub.core.context import tuning_context
from embedl_hub.tracking import log_metric, log_param

__all__ = ['tuning_context', 'log_metric', 'log_param']

try:
    __version__ = importlib.metadata.version("embedl_hub")
except importlib.metadata.PackageNotFoundError:
    __version__ = "0.0.0"



# Copyright (C) 2025 Embedl AB

"""
embedl-hub export - send torch script model to Qualcomm AI Hub and retrieve an onnx model.
"""

from pathlib import Path

import typer

# All other embedl_hub imports should be done inside the function.
from embedl_hub.cli.helper import (
    DEVICE_HELPER,
    OUTPUT_FILE_HELPER,
    SIZE_HELPER,
)

export_cli = typer.Typer(
    invoke_without_command=True,
    no_args_is_help=True,
)


@export_cli.command("export")
def export_command(
    model: Path = typer.Option(
        ...,
        "-m",
        "--model",
        help="Path to the TorchScript model file to be exported.",
        show_default=False,
    ),
    size: str = typer.Option(
        ...,
        "--size",
        "-s",
        help=SIZE_HELPER,
        show_default=False,
    ),
    device: str = typer.Option(
        ...,
        "-d",
        "--device",
        help=DEVICE_HELPER
        + " Exporting for a specific device can improve compatibility.",
        show_default=False,
    ),
    output_file: str = typer.Option(
        None,
        "-o",
        "--output-file",
        help=OUTPUT_FILE_HELPER,
        show_default=False,
    ),
):
    """
    Compile a TorchScript model into an ONNX model using Qualcomm AI Hub.
    Qualcomm AI Hub may return a zip file containing multiple files.

    Required arguments:
        --model
        --size
        --device

    Examples
    --------
    Export the TorchScript model `tuned_model.pt` with input size 224x224
    for the Samsung Galaxy S24:

        $ embedl-hub export -m tuned_model.pt --size 224,224 --device "Samsung Galaxy S24"

    Export the TorchScript model `tuned_model.pt` with input size 224x224
    for the Samsung Galaxy S24, and save it to `./my_outputs/model.onnx`:

        $ embedl-hub export -m tuned_model.pt  --size 224,224 --device "Samsung Galaxy S24" -o ./my_outputs/model.onnx

    """

    # pylint: disable=import-outside-toplevel
    from embedl_hub.cli.utils import prepare_image_size
    from embedl_hub.core.compile import CompileError, export_model
    from embedl_hub.core.hub_logging import console
    # pylint: enable=import-outside-toplevel

    if not model:
        raise ValueError("Please specify a model to export using --model")
    if not size:
        raise ValueError(
            "Please specify input image size using --size, e.g. 224,224"
        )
    if not output_file:
        output_file = model.with_suffix(".onnx").as_posix()
        console.print(
            f"[yellow]No output file specified, using {output_file}[/]"
        )
    image_size = prepare_image_size(size)
    try:
        export_model(
            model_file=model,
            device=device,
            output_file=output_file,
            image_size=image_size,
        )
        console.print("[green]✓ Exported model to ONNX[/]")

        # TODO: upload artifacts / metrics to web
    except (CompileError, ValueError) as error:
        console.print(f"[red]✗ {error}[/]")
        raise typer.Exit(1)



# Copyright (C) 2025 Embedl AB

"""Load a model from the Embedl Hub."""

from enum import StrEnum

from timm import create_model, list_pretrained
from torch import nn
from torchvision.models import get_model, get_model_weights


def _change_num_classes_linear(linear_layer, num_classes: int):
    """Change the number of classes in a linear layer."""
    return nn.Linear(
        in_features=linear_layer.in_features,
        out_features=num_classes,
        bias=linear_layer.bias is not None,
    )


def change_num_classes_linear(
    layer_name: str, model: nn.Module, num_classes: int
):
    """Change the number of classes in a linear layer."""
    linear_layer = getattr(model, layer_name)
    new_linear_layer = _change_num_classes_linear(linear_layer, num_classes)
    setattr(model, layer_name, new_linear_layer)


def change_num_classes_sequential(
    layer_name: str, model: nn.Module, num_classes: int
):
    """Change the number of classes in a sequential layer."""

    def _recursive_search_for_linear(layer: nn.Module, num_classes: int):
        if isinstance(layer, nn.Linear):
            sequential_layer[-1] = _change_num_classes_linear(
                layer, num_classes
            )
        elif isinstance(layer, nn.Sequential):
            _recursive_search_for_linear(sequential_layer[-1], num_classes)

    sequential_layer = getattr(model, layer_name)
    _recursive_search_for_linear(sequential_layer, num_classes)


class ClassifierName(StrEnum):
    """Enum for classifier names in torchvision models."""

    CLASSIFIER = "classifier"
    FC = "fc"
    HEAD = "head"
    HEADS = "heads"


# Map model names to their classifier layer type
model_to_classifier_type = {
    "alexnet": ClassifierName.CLASSIFIER,
    "convnext_base": ClassifierName.CLASSIFIER,
    "convnext_large": ClassifierName.CLASSIFIER,
    "convnext_small": ClassifierName.CLASSIFIER,
    "convnext_tiny": ClassifierName.CLASSIFIER,
    "densenet101": ClassifierName.CLASSIFIER,
    "densenet121": ClassifierName.CLASSIFIER,
    "densenet161": ClassifierName.CLASSIFIER,
    "densenet169": ClassifierName.CLASSIFIER,
    "densenet201": ClassifierName.CLASSIFIER,
    "efficientnet_b0": ClassifierName.CLASSIFIER,
    "efficientnet_b1": ClassifierName.CLASSIFIER,
    "efficientnet_b2": ClassifierName.CLASSIFIER,
    "efficientnet_b3": ClassifierName.CLASSIFIER,
    "efficientnet_b4": ClassifierName.CLASSIFIER,
    "efficientnet_b5": ClassifierName.CLASSIFIER,
    "efficientnet_b6": ClassifierName.CLASSIFIER,
    "efficientnet_b7": ClassifierName.CLASSIFIER,
    "efficientnet_v2_l": ClassifierName.CLASSIFIER,
    "efficientnet_v2_m": ClassifierName.CLASSIFIER,
    "efficientnet_v2_s": ClassifierName.CLASSIFIER,
    "googlenet": ClassifierName.FC,
    "inception_v3": ClassifierName.FC,
    "maxvit_t": ClassifierName.CLASSIFIER,
    "mnasnet0_5": ClassifierName.CLASSIFIER,
    "mnasnet0_75": ClassifierName.CLASSIFIER,
    "mnasnet1_0": ClassifierName.CLASSIFIER,
    "mnasnet1_3": ClassifierName.CLASSIFIER,
    "mobilenet_v2": ClassifierName.CLASSIFIER,
    "mobilenet_v3_large": ClassifierName.CLASSIFIER,
    "mobilenet_v3_small": ClassifierName.CLASSIFIER,
    "regnet_x_16gf": ClassifierName.FC,
    "regnet_x_1_6gf": ClassifierName.FC,
    "regnet_x_32gf": ClassifierName.FC,
    "regnet_x_3_2gf": ClassifierName.FC,
    "regnet_x_400mf": ClassifierName.FC,
    "regnet_x_800mf": ClassifierName.FC,
    "regnet_x_8gf": ClassifierName.FC,
    "regnet_y_128gf": ClassifierName.FC,
    "regnet_y_16gf": ClassifierName.FC,
    "regnet_y_1_6gf": ClassifierName.FC,
    "regnet_y_32gf": ClassifierName.FC,
    "regnet_y_3_2gf": ClassifierName.FC,
    "regnet_y_400mf": ClassifierName.FC,
    "regnet_y_800mf": ClassifierName.FC,
    "regnet_y_8gf": ClassifierName.FC,
    "resnet101": ClassifierName.FC,
    "resnet152": ClassifierName.FC,
    "resnet18": ClassifierName.FC,
    "resnet34": ClassifierName.FC,
    "resnet50": ClassifierName.FC,
    "resnext101_32x8d": ClassifierName.FC,
    "resnext101_64x4d": ClassifierName.FC,
    "resnext50_32x4d": ClassifierName.FC,
    "shufflenet_v2_x0_5": ClassifierName.FC,
    "shufflenet_v2_x1_0": ClassifierName.FC,
    "shufflenet_v2_x1_5": ClassifierName.FC,
    "shufflenet_v2_x2_0": ClassifierName.FC,
    "swin_b": ClassifierName.HEAD,
    "swin_s": ClassifierName.HEAD,
    "swin_t": ClassifierName.HEAD,
    "swin_v2_b": ClassifierName.HEAD,
    "swin_v2_s": ClassifierName.HEAD,
    "swin_v2_t": ClassifierName.HEAD,
    "vgg11": ClassifierName.CLASSIFIER,
    "vgg11_bn": ClassifierName.CLASSIFIER,
    "vgg13": ClassifierName.CLASSIFIER,
    "vgg13_bn": ClassifierName.CLASSIFIER,
    "vgg16": ClassifierName.CLASSIFIER,
    "vgg16_bn": ClassifierName.CLASSIFIER,
    "vgg19": ClassifierName.CLASSIFIER,
    "vgg19_bn": ClassifierName.CLASSIFIER,
    "vit_b_16": ClassifierName.HEADS,
    "vit_b_32": ClassifierName.HEADS,
    "vit_h_14": ClassifierName.HEADS,
    "vit_l_16": ClassifierName.HEADS,
    "vit_l_32": ClassifierName.HEADS,
    "wide_resnet101_2": ClassifierName.FC,
    "wide_resnet50_2": ClassifierName.FC,
}


def _load_torchvision_model(clean_name: str, pre_trained: bool) -> nn.Module:
    """Create a torchvision model."""
    weights_enum = get_model_weights(clean_name)
    if pre_trained:
        weights = weights_enum.DEFAULT
    else:
        weights = None
    model = get_model(clean_name, weights=weights)
    model.eval()
    return model


def clean_model_id(model_id: str) -> str:
    """
    Clean the model_id to get the model name.

    1. Remove the source prefix (e.g., "torchvision-")
    2. Remove the quantization suffix (e.g., "-int8")
    """

    source_prefixes = ("torchvision-", "timm-")
    quantization_suffixes = ("-int8", "-fp16", "-fp32", "-mixed")

    if not any(model_id.startswith(prefix) for prefix in source_prefixes):
        raise ValueError("Model ID must contain a valid source prefix.")
    if not any(model_id.endswith(suffix) for suffix in quantization_suffixes):
        raise ValueError("Model ID must contain a valid quantization suffix.")

    model_id_components = model_id.split("-")[
        1:-1
    ]  # Remove source prefix and suffix
    return "_".join(model_id_components).lower()


def load_model_from_torchvision(model_id: str, pre_trained: bool) -> nn.Module:
    """Load a model from torchvision.

    Args:
        model_id (str): The ID of the model to load.
        pre_trained (bool): Whether to load the pre-trained weights.
    """

    clean_name = clean_model_id(model_id)
    return _load_torchvision_model(clean_name, pre_trained)


def change_num_classes(
    model_id: str, model: nn.Module, num_classes: int
) -> nn.Module:
    """Change the number of classes in the model.

    Args:
        model_id (str): The ID of the model.
        model (nn.Module): The model to modify.
        num_classes (int): The new number of classes.
    """
    try:
        clean_name = clean_model_id(model_id)
        classifier_type = model_to_classifier_type[clean_name].value
    except KeyError as e:
        raise ValueError(
            f"Model {model_id} is currently not supported for fine tuning. "
        ) from e
    last_layer = getattr(model, classifier_type)
    if isinstance(last_layer, nn.Linear):
        class_changer = change_num_classes_linear
    elif isinstance(last_layer, nn.Sequential):
        class_changer = change_num_classes_sequential
    else:
        raise ValueError(
            f"Unsupported classifier type {type(last_layer)} for model {model_id}."
        )
    class_changer(classifier_type, model, num_classes)
    return model


def load_torchvision_with_num_classes(
    model_id: str, pre_trained: bool, num_classes: int
) -> nn.Module:
    """Load a torchvision model with a specific number of classes."""
    model = load_model_from_torchvision(model_id, pre_trained)
    return change_num_classes(model_id, model, num_classes)


def model_id_to_timm_name(model_id: str) -> str:
    """Find a timm model with a specific ID."""
    clean_name = clean_model_id(model_id)
    for timm_name in list_pretrained():
        if timm_name.replace(".", "_").casefold() == clean_name.casefold():
            return timm_name
    raise ValueError(f"Model {model_id} not found in timm models.")


def load_timm_with_num_classes(
    model_id: str, pre_trained: bool, num_classes: int
) -> nn.Module:
    """Load a timm model with a specific number of classes."""

    model_name = model_id_to_timm_name(model_id)
    model = create_model(
        model_name, pretrained=pre_trained, num_classes=num_classes
    )
    return model


def load_model_with_num_classes(
    model_id: str, pre_trained: bool, num_classes: int
) -> nn.Module:
    """Load a model with a specific number of classes."""
    if model_id.startswith("torchvision-"):
        return load_torchvision_with_num_classes(
            model_id, pre_trained, num_classes
        )
    if model_id.startswith("timm-"):
        return load_timm_with_num_classes(model_id, pre_trained, num_classes)
    raise ValueError(
        f"Model {model_id} is currently not supported for fine tuning. "
        "Only torchvision and timm models are supported."
    )



# Copyright (C) 2025 Embedl AB

"""Tune batch size for a model."""

from typing import Optional

import pytorch_lightning as pl
import torch

from embedl_hub.core.tuning.tuning_config import TuningConfig


def tune_batch_size(
    config: TuningConfig,
    trainer: pl.Trainer,
    hub_data_module: pl.LightningDataModule,
    hub_module: pl.LightningModule,
) -> None:
    """Find maximum batch size for a model."""

    if not torch.cuda.is_available():
        raise ValueError(
            "CUDA is not available. Batch size tuning is currently not supported on CPU."
        )

    tuner = pl.tuner.tuning.Tuner(trainer)
    auto_batch_size: Optional[int] = tuner.scale_batch_size(
        datamodule=hub_data_module,
        model=hub_module,
    )
    if auto_batch_size is None:
        raise ValueError("Auto batch size tuning failed.")
    config.batch_size = auto_batch_size


def maybe_tune_batch_size(
    config: TuningConfig,
    trainer: pl.Trainer,
    hub_data_module: pl.LightningDataModule,
    hub_module: pl.LightningModule,
) -> None:
    """Tune batch size if set to auto."""
    if config.batch_size == "auto":
        tune_batch_size(config, trainer, hub_data_module, hub_module)



# Copyright (C) 2025 Embedl AB

"""CLI component for fine-tuning models."""

from pathlib import Path

import typer

# All other embedl_hub imports should be done inside the function.
from embedl_hub.cli.helper import CONFIG_HELPER, SIZE_HELPER

tune_cli = typer.Typer()
__all__ = ["tune_cli"]


@tune_cli.command("tune")
def tune_command(
    model_id: str = typer.Option(
        None,
        "--model_id",
        "--id",
        "-i",
        help=(
            "Model ID to fine-tune. "
            "Browse available models at https://hub.embedl.com/browse. (required)"
        ),
        show_default=False,
    ),
    num_classes: int = typer.Option(
        None,
        "--num-classes",
        "-nc",
        help="Number of output classes for the classification task. (required)",
        show_default=False,
    ),
    data_path: str = typer.Option(
        None,
        "--data",
        "-d",
        help=(
            "Path to the dataset used for tuning. "
            "Supports a modified torchvision ImageFolder format. "
            "Use a single directory (e.g., /path/to/dataset) for a training-only "
            "set, or include '/train' and '/val' subdirectories for separate "
            "training and validation splits. Each subdirectory must contain one "
            "folder per class with images. (required)"
        ),
        show_default=False,
    ),
    batch_size: int = typer.Option(
        None,
        "--batch-size",
        "-b",
        help=(
            "Batch size used during training. "
            "If unset, will auto-tune to the maximum that fits in GPU memory. "
            "Auto-tuning is unavailable without a GPU, so on CPU-only systems "
            "you must set this manually."
        ),
        show_default=False,
    ),
    max_learning_rate: float = typer.Option(
        None,
        "--learning-rate",
        "-lr",
        help=(
            "Maximum learning rate of the one-cycle scheduler during optimization. "
            "Default, search for optimal learning rate."
        ),
        show_default=False,
    ),
    size: str = typer.Option(
        None,
        "--size",
        help=SIZE_HELPER,
        show_default=False,
    ),
    epochs: int = typer.Option(
        None,
        "--epochs",
        "-e",
        help="Number of training epochs. (required)",
        show_default=False,
    ),
    config_path: Path = typer.Option(
        None,
        "--config",
        "-c",
        help=CONFIG_HELPER,
    ),
):
    """
    Fine-tune a model on your dataset.

    Required arguments (if not provided through --config):
        --model-id
        --num-classes
        --data
        --epochs

    Examples:
    ---------
    Tune the torchvision AlexNet model for 10 classes using the dataset
    at `/path/to/dataset` for 10 epochs:

        $ embedl-hub tune -i torchvision/alexnet -nc 10 -d /path/to/dataset -e 10

    Tune the torchvision ResNet50 model for 10 classes using the dataset
    at `/path/to/dataset` with custom settings from `config.yaml`:

        $ embedl-hub tune -i torchvision/resnet50 -nc 10 -d /path/to/dataset -c /path/to/config.yaml

    """

    # pylint: disable=import-outside-toplevel
    from embedl_hub.cli.utils import assert_api_config, remove_none_values
    from embedl_hub.core.config import load_default_config_with_size
    from embedl_hub.core.context import require_embedl_hub_context
    from embedl_hub.core.hub_logging import console
    from embedl_hub.core.tuning.tuner import tune_model
    from embedl_hub.core.tuning.tuning_config import TuningConfig
    # pylint: enable=import-outside-toplevel

    assert_api_config()
    ctx = require_embedl_hub_context()

    cfg = load_default_config_with_size(TuningConfig, size, "tune")
    cli_flags = {
        "model_id": model_id,
        "num_classes": num_classes,
        "data_path": data_path,
        "batch_size": batch_size,
        "max_learning_rate": max_learning_rate,
        "epochs": epochs,
    }
    cfg = cfg.merge_yaml(other=config_path, **remove_none_values(cli_flags))
    try:
        cfg.validate_config()
    except ValueError as e:
        console.print(f"[red]Error: {e}[/]")
        raise typer.Exit(1)

    tune_model(
        config=cfg,
        project_name=ctx["project_name"],
        experiment_name=ctx["experiment_name"],
    )



# Copyright (C) 2025 Embedl AB

"""Tune learning rate for a model."""

import pytorch_lightning as pl

from embedl_hub.core.tuning.tuning_config import TuningConfig


def tune_learning_rate(
    config: TuningConfig,
    trainer: pl.Trainer,
    hub_data_module: pl.LightningDataModule,
    hub_module: pl.LightningModule,
):
    """Automatically find learning rate for model."""
    # Tune trainer
    tuner = pl.tuner.tuning.Tuner(trainer)
    lr_finder = tuner.lr_find(
        min_lr=1e-5,
        train_dataloaders=hub_data_module.train_dataloader(),
        model=hub_module,
    )
    if lr_finder is None:
        raise ValueError("LR finder failed.")
    suggestion = lr_finder.suggestion()
    if suggestion is None:
        raise ValueError(
            "LR finder did not return a valid learning rate suggestion."
        )
    config.max_learning_rate = suggestion


def maybe_tune_learning_rate(
    config: TuningConfig,
    trainer: pl.Trainer,
    hub_data_module: pl.LightningDataModule,
    hub_module: pl.LightningModule,
) -> None:
    """Tune learning rate if set to auto."""
    if config.max_learning_rate == "auto":
        tune_learning_rate(config, trainer, hub_data_module, hub_module)
